from guillotina.component import get_utility
from guillotina_nats.interfaces import INatsUtility

import pytest


@pytest.mark.asyncio
async def test_connection(natsd, container_requester):
    async with container_requester as requester:  # noqa
        nats = get_utility(INatsUtility)

        assert nats._initialized



# -*- coding: utf-8 -*-
from nats.aio.client import Client as NATS
from nats.aio.errors import ErrConnectionClosed
from nats.aio.errors import ErrConnectionReconnecting
from nats.aio.errors import ErrNoServers
from nats.aio.errors import ErrTimeout
from stan.aio.client import Client as STAN
from stan.aio.errors import StanError

import asyncio
import logging
import os
import uuid

logger = logging.getLogger("guillotina_nats")


# Configuration Utility


class NatsUtility(object):
    def __init__(self, settings, loop=None):
        self._loop = loop
        self._hosts = settings["hosts"]
        self._timeout = float(settings["timeout"])
        self._subscriptions = []
        self._stream_subscriptions = []
        self._stan = settings.get("stan", None)
        self._name = settings.get("name", None)
        self._thread = settings.get("thread", False)
        self._uuid = os.environ.get("HOSTNAME", uuid.uuid4().hex)
        self._initialized = False
        self.lock = asyncio.Lock()
        self.sc = None
        self.nc = None

    async def subscribe(self, handler, key, group=""):
        if self.nc.is_connected:
            sid = await self.nc.subscribe(key, queue=group, cb=handler)
            self._subscriptions.append(sid)
            logger.info("Subscribed to " + key)
            return sid
        else:
            raise ErrConnectionClosed("Could not subscribe")

    async def unsubscribe(self, sid):
        if self.nc.is_connected and sid is not None:
            await self.nc.unsubscribe(sid)
            self._subscriptions.remove(sid)
        else:
            raise ErrConnectionClosed("Could not unsubscribe")

    async def stream_subscribe(self, handler, key, **params):
        if self.sc._conn_id:
            sid = await self.sc.subscribe(key, cb=handler, **params)
            self._stream_subscriptions.append(sid)
            logger.info("Subscribed to " + key)
            return sid
        else:
            raise ErrConnectionClosed("Could not subscribe")

    async def stream_unsubscribe(self, sid):
        if self.sc._conn_id:
            await sid.unsubscribe()
            self._stream_subscriptions.remove(sid)
        else:
            raise ErrConnectionClosed("Could not unsubscribe")

    async def publish(self, key, value):
        if self.nc.is_connected:
            await self.nc.publish(key, value)
        else:
            raise ErrConnectionClosed("Could not publish")

    async def stream_publish(self, key, value):
        async def cb(ack):
            pass

        try:
            await self.sc.publish(key, value, ack_handler=cb, ack_wait=60)
        except AttributeError:
            await self.sc.connect(self._stan, self._uuid, nats=self.nc)
            await asyncio.sleep(2)
            await self.sc.publish(key, value, ack_handler=cb, ack_wait=60)

    async def request(self, key, value, timeout=None):
        if timeout is None:
            timeout = self._timeout
        if self.nc.is_connected:
            try:
                return await self.nc.request(key, value, timeout)
            except ErrTimeout:
                return
        else:
            raise ErrConnectionClosed("Could not subscribe")

    async def stream(self, channel_name, data):
        if self.sc._conn_id:
            await self.sc.publish(channel_name, data)
        else:
            raise ErrConnectionClosed("Could not publish")

    async def initialized(self):
        if self._initialized:
            return True
        async with self.lock:
            return True
        return False

    async def initialize(self, app=None):
        # No asyncio loop to run
        async with self.lock:
            self.nc = NATS()
            options = {
                "servers": self._hosts,
                "loop": self._loop,
                "disconnected_cb": self.disconnected_cb,
                "reconnected_cb": self.reconnected_cb,
                "error_cb": self.error_cb,
                "closed_cb": self.closed_cb,
                "name": self._name,
                "verbose": True,
            }

            try:
                await self.nc.connect(**options)
            except ErrNoServers:
                logger.exception("No servers found")
                raise

            logger.info("Connected to nats")

            if self._stan is not None:
                self.sc = STAN()
                await self.sc.connect(self._stan, self._uuid, nats=self.nc)
                logger.info("Connected to stan")
        self._initialized = True

    async def finalize(self, app):
        if self.sc:
            for sid in self._stream_subscriptions:
                await sid.unsubscribe()
            try:
                await self.sc.close()
            except ErrTimeout:
                pass
            except StanError:
                pass
        if self.nc:
            try:
                await self.nc.flush()
            except RuntimeError:
                pass
            except AttributeError:
                pass
            for key in self._subscriptions:
                await self.nc.unsubscribe(key)
            try:
                await self.nc.drain()
            except AttributeError:
                pass
            except ErrConnectionReconnecting:
                pass
            try:
                await self.nc.close()
            except RuntimeError:
                pass

    async def disconnected_cb(self):
        logger.info("Got disconnected!")

    async def reconnected_cb(self):
        # See who we are connected to on reconnect.
        logger.info("Got reconnected to {url}".format(url=self.nc.connected_url.netloc))

    async def error_cb(self, e):
        logger.info("There was an error: {}".format(e))

    async def closed_cb(self):
        logger.info("Connection is closed")



from guillotina import testing
from guillotina_nats.tests.utils import Gnatsd
from guillotina_nats.tests.utils import StanServer
from guillotina_nats.tests.utils import start_gnatsd
from guillotina_nats.tests.utils import start_nats_streaming
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile

import os
import platform
import pytest


def base_settings_configurator(settings):
    if "applications" in settings:
        settings["applications"].append("guillotina_nats")
    else:
        settings["applications"] = ["guillotina_nats"]

    settings["load_utilities"]["nats"] = {
        "provides": "guillotina_nats.interfaces.INatsUtility",
        "factory": "guillotina_nats.utility.NatsUtility",
        "settings": {"timeout": 0.2, "hosts": ["nats://localhost:4222"], "stan": None},
    }


testing.configure_with(base_settings_configurator)


@pytest.fixture(scope="function")
def natsd():
    if not os.path.isfile("nats-server"):
        version = "v2.0.0"
        arch = platform.machine()
        if arch == "x86_64":
            arch = "amd64"
        system = platform.system().lower()

        url = f"https://github.com/nats-io/nats-server/releases/download/{version}/nats-server-{version}-{system}-{arch}.zip"

        resp = urlopen(url)
        zipfile = ZipFile(BytesIO(resp.read()))

        file = zipfile.open(f"nats-server-{version}-{system}-{arch}/nats-server")
        content = file.read()
        with open("nats-server", "wb") as f:
            f.write(content)
        os.chmod("nats-server", 755)

    server = Gnatsd(port=4222)
    server.bin_name = "nats-server"
    server.path = os.getcwd()
    start_gnatsd(server)
    print("Started natsd")
    yield
    server.stop()


@pytest.fixture(scope="function")
def stand():
    if not os.path.isfile("nats-streaming-server"):
        version = "v0.16.2"
        arch = platform.machine()
        if arch == "x86_64":
            arch = "amd64"
        system = platform.system().lower()

        url = f"https://github.com/nats-io/nats-streaming-server/releases/download/{version}/nats-streaming-server-{version}-{system}-{arch}.zip"

        resp = urlopen(url)
        zipfile = ZipFile(BytesIO(resp.read()))

        file = zipfile.open(
            f"nats-streaming-server-{version}-{system}-{arch}/nats-streaming-server"
        )
        content = file.read()
        with open("nats-streaming-server", "wb") as f:
            f.write(content)
        os.chmod("nats-streaming-server", 755)

    server = StanServer(port=4222)
    server.path = os.getcwd()
    start_nats_streaming(server)
    print("Started stand")
    yield
    server.stop()



from guillotina.component import get_utility
from guillotina_nats.interfaces import INatsUtility

import asyncio
import pytest


@pytest.mark.asyncio
async def test_pubsub(natsd, container_requester):
    async with container_requester as requester:  # noqa
        nats = get_utility(INatsUtility)

        variable = []

        async def callback(msg):
            variable.append("done")

        await nats.subscribe(callback, "testing")
        await nats.publish("testing", b"hola")

        await asyncio.sleep(1)
        assert variable[0] == "done"



from guillotina.component import get_utility
from guillotina_nats.interfaces import INatsUtility

import asyncio
import pytest


@pytest.mark.asyncio
@pytest.mark.app_settings(
    {"load_utilities": {"nats": {"settings": {"stan": "test-cluster"}}}}
)
async def test_stream(stand, container_requester):
    async with container_requester as requester:  # noqa
        nats = get_utility(INatsUtility)

        variable = []

        await nats.stream("testingst", b"hola")

        async def callback(msg):
            variable.append("done")

        await nats.stream_subscribe(callback, "testingst", start_at="first")

        await asyncio.sleep(1)

        assert variable[0] == "done"



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup

import re


def load_reqs(filename):
    with open(filename) as reqs_file:
        return [
            re.sub('==', '>=', line) for line in reqs_file.readlines()
            if not (re.match('\s*#', line) or re.match('-e', line) or re.match('-r', line))
        ]


requirements = load_reqs('requirements.txt')
test_requirements = load_reqs('requirements-test.txt')

setup(
    name='guillotina_nats',
    version=open('VERSION').read().strip(),
    long_description=(open('README.rst').read() + '\n' +
                      open('CHANGELOG.rst').read()),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    url='https://github.com/guillotinaweb/guillotina_nats',
    license='BSD',
    author="Ramon Navarro Bosch",
    author_email="ramon@flaps.io",
    setup_requires=[
        'pytest-runner',
    ],
    zip_safe=True,
    include_package_data=True,
    packages=find_packages(),
    install_requires=requirements,
    tests_require=test_requirements
)



import http.client
import json
import os
import random
import signal
import subprocess
import time


class Gnatsd(object):
    def __init__(
        self,
        port=4222,
        user="",
        password="",
        token="",
        timeout=0,
        http_port=8222,
        debug=False,
        tls=False,
        cluster_listen=None,
        routes=[],
        config_file=None,
    ):
        self.port = port
        self.user = user
        self.password = password
        self.timeout = timeout
        self.http_port = http_port
        self.proc = None
        self.debug = debug
        self.tls = tls
        self.token = token
        self.cluster_listen = cluster_listen
        self.routes = routes
        self.bin_name = "gnatsd"
        self.config_file = config_file

        env_debug_flag = os.environ.get("DEBUG_NATS_TEST")
        if env_debug_flag == "true":
            self.debug = True

    def start(self):
        cmd = [
            f"{self.path}/{self.bin_name}",
            "-p",
            "%d" % self.port,
            "-m",
            "%d" % self.http_port,
            "-a",
            "127.0.0.1",
        ]
        if self.user != "":
            cmd.append("--user")
            cmd.append(self.user)
        if self.password != "":
            cmd.append("--pass")
            cmd.append(self.password)

        if self.token != "":
            cmd.append("--auth")
            cmd.append(self.token)

        if self.debug:
            cmd.append("-DV")

        if self.tls:
            cmd.append("--tls")
            cmd.append("--tlscert")
            cmd.append("tests/certs/server-cert.pem")
            cmd.append("--tlskey")
            cmd.append("tests/certs/server-key.pem")
            cmd.append("--tlsverify")
            cmd.append("--tlscacert")
            cmd.append("tests/certs/ca.pem")

        if self.cluster_listen is not None:
            cmd.append("--cluster_listen")
            cmd.append(self.cluster_listen)

        if len(self.routes) > 0:
            cmd.append("--routes")
            cmd.append(",".join(self.routes))

        if self.config_file is not None:
            cmd.append("--config")
            cmd.append(self.config_file)

        if self.debug:
            self.proc = subprocess.Popen(cmd)
        else:
            self.proc = subprocess.Popen(
                cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )

        if self.debug:
            if self.proc is None:
                print(
                    "[\031[0;33mDEBUG\033[0;0m] Failed to start server listening on port %d started."
                    % self.port
                )
            else:
                print(
                    "[\033[0;33mDEBUG\033[0;0m] Server listening on port %d started."
                    % self.port
                )
        return self.proc

    def stop(self):
        if self.debug:
            print(
                "[\033[0;33mDEBUG\033[0;0m] Server listening on %d will stop."
                % self.port
            )

        if self.debug:
            if self.proc is None:
                print(
                    "[\033[0;31mDEBUG\033[0;0m] Failed terminating server listening on port %d"
                    % self.port
                )

        if self.proc.returncode is not None:
            if self.debug:
                print(
                    "[\033[0;31mDEBUG\033[0;0m] Server listening on port {port} finished running already with exit {ret}".format(
                        port=self.port, ret=self.proc.returncode
                    )
                )
        else:
            os.kill(self.proc.pid, signal.SIGKILL)
            self.proc.wait()
            if self.debug:
                print(
                    "[\033[0;33mDEBUG\033[0;0m] Server listening on %d was stopped."
                    % self.port
                )


class NatsServer(Gnatsd):
    def __init__(self):
        super(Gnatsd, self)
        self.bin_name = "nats-server"


def start_gnatsd(gnatsd: Gnatsd):
    gnatsd.start()

    endpoint = "127.0.0.1:{port}".format(port=gnatsd.http_port)
    retries = 0
    while True:
        if retries > 100:
            break

        try:
            httpclient = http.client.HTTPConnection(endpoint, timeout=5)
            httpclient.request("GET", "/varz")
            response = httpclient.getresponse()
            if response.code == 200:
                break
        except:
            retries += 1
            time.sleep(0.1)


class StanServer(object):
    def __init__(
        self,
        port=4222,
        user="",
        password="",
        token="",
        timeout=0,
        http_port=8222,
        debug=False,
        tls=False,
        cluster_listen=None,
        routes=[],
    ):
        self.port = port
        self.user = user
        self.password = password
        self.timeout = timeout
        self.http_port = http_port
        self.proc = None
        self.debug = debug
        self.tls = tls
        self.token = token
        self.cluster_listen = cluster_listen
        self.routes = routes

        env_debug_flag = os.environ.get("DEBUG_NATS_TEST")
        if env_debug_flag == "true":
            self.debug = True

    def start(self):
        cmd = [
            f"{self.path}/nats-streaming-server",
            "-p",
            "%d" % self.port,
            "-m",
            "%d" % self.http_port,
            "-a",
            "127.0.0.1",
        ]
        if self.user != "":
            cmd.append("--user")
            cmd.append(self.user)
        if self.password != "":
            cmd.append("--pass")
            cmd.append(self.password)

        if self.token != "":
            cmd.append("--auth")
            cmd.append(self.token)

        if self.debug:
            cmd.append("-SDV")
            cmd.append("-DV")

        if self.cluster_listen is not None:
            cmd.append("--cluster_listen")
            cmd.append(self.cluster_listen)

        if len(self.routes) > 0:
            cmd.append("--routes")
            cmd.append(",".join(self.routes))

        if self.debug:
            self.proc = subprocess.Popen(cmd)
        else:
            self.proc = subprocess.Popen(
                cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )

        if self.debug:
            if self.proc is None:
                print(
                    "[\031[0;33mDEBUG\033[0;0m] Failed to start server listening on port %d started."
                    % self.port
                )
            else:
                print(
                    "[\033[0;33mDEBUG\033[0;0m] Server listening on port %d started."
                    % self.port
                )
        return self.proc

    def stop(self):
        if self.debug:
            print(
                "[\033[0;33mDEBUG\033[0;0m] Server listening on %d will stop."
                % self.port
            )

        if self.debug:
            if self.proc is None:
                print(
                    "[\033[0;31mDEBUG\033[0;0m] Failed terminating server listening on port %d"
                    % self.port
                )

        if self.proc.returncode is not None:
            if self.debug:
                print(
                    "[\033[0;31mDEBUG\033[0;0m] Server listening on port {port} finished running already with exit {ret}".format(
                        port=self.port, ret=self.proc.returncode
                    )
                )
        else:
            os.kill(self.proc.pid, signal.SIGKILL)
            self.proc.wait()
            if self.debug:
                print(
                    "[\033[0;33mDEBUG\033[0;0m] Server listening on %d was stopped."
                    % self.port
                )


def start_nats_streaming(server: StanServer):
    server.start()

    endpoint = "127.0.0.1:{port}".format(port=server.http_port)
    retries = 0
    while True:
        if retries > 100:
            break

        try:
            httpclient = http.client.HTTPConnection(endpoint, timeout=5)
            httpclient.request("GET", "/subsz")
            response = httpclient.getresponse()
            if response.code == 200:
                # Check that at least the minimum subjects for a
                # NATS Streaming session are ready.
                connz = json.loads(response.read())
                if connz["num_subscriptions"] < 7:
                    continue
                else:
                    break
        except:
            retries += 1
            time.sleep(0.1)


def generate_client_id():
    return "%x" % random.SystemRandom().getrandbits(0x58)
